#!/usr/bin/env python

"""
Generic file access.

Copyright (C) 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.data import vint
import bz2, zlib

# Constants.

WRITE_CACHE_SIZE  = 100000
READ_CACHE_SIZE   = 10000
READ_CACHE_RESIZE = 5000

compressors = [("b", bz2.compress), ("z", zlib.compress)]
decompressors = {"b" : bz2.decompress, "z" : zlib.decompress}

class File:

    "A basic file abstraction."

    def __init__(self, f):
        self.f = f
        self.reset()

    def reset(self):

        "To be used to reset the state of the reader or writer between records."

        pass

    def rewind(self):
        self.seek(0)
        self.reset()

    def seek(self, offset):

        "To be defined by readers."

        pass

    def flush(self):

        "To be defined by writers."

        pass

    def close(self):
        if self.f is not None:
            self.flush()
            self.f.close()
            self.f = None

class FileWriter(File):

    "Writing basic data types to files."

    def __init__(self, f):
        File.__init__(self, f)
        self.cache = []
        self.cache_length = 0

    def write_number(self, number):

        "Write 'number' to the file using a variable length encoding."

        self.write(vint(number))

    def write_string(self, s, compress=0):

        """
        Write 's' to the file, recording its length and compressing the string
        if 'compress' is set to a true value.
        """

        # Convert Unicode objects to strings.

        if isinstance(s, unicode):
            s = s.encode("utf-8")

        # Compress the string if requested.

        if compress:
            for flag, fn in compressors:
                cs = fn(s)

                # Take the first string shorter than the original.

                if len(cs) < len(s):
                    s = cs
                    break
            else:
                flag = "-"

        else:
            flag = ""

        # Write the length of the data before the data itself.

        length = len(s)
        self.write(flag + vint(length) + s)

    # Cache-affected methods.

    def write(self, s):
        self.cache.append(s)
        self.cache_length += len(s)
        if self.cache_length >= WRITE_CACHE_SIZE:
            self.flush()

    def tell(self):
        return self.f.tell() + self.cache_length

    def flush(self):
        self.f.write("".join(self.cache))
        self.cache = []
        self.cache_length = 0

class FileReader(File):

    "Reading basic data types from files."

    def __init__(self, f):
        File.__init__(self, f)
        self.reset_cache()

    def reset_cache(self):
        self.cache = ""
        self.cache_length = 0
        self.cache_start = 0

    def read_number(self):

        "Read a number from the file."

        # Read each byte, adding it to the number.

        shift = 0
        number = 0
        read = self.read

        try:
            csd = ord(read(1))
            while csd & 128:
                number += ((csd & 127) << shift)
                shift += 7
                csd = ord(read(1))
            else:
                number += (csd << shift)
        except TypeError:
            raise EOFError

        return number

    def read_string(self, decompress=0):

        """
        Read a string from the file, decompressing the stored data if
        'decompress' is set to a true value.
        """

        # Decompress the data if requested.

        if decompress:
            flag = self.read(1)
        else:
            flag = "-"

        length = self.read_number()
        s = self.read(length)

        # Perform decompression if applicable.

        if flag != "-":
            fn = decompressors[flag]
            s = fn(s)

        # Convert strings to Unicode objects.

        return unicode(s, "utf-8")

    # Cache-affected methods.

    def read(self, n):
        needed = n - (self.cache_length - self.cache_start)

        # Read the needed number of characters, if possible.

        if needed > 0:
            s = self.f.read(max(needed, READ_CACHE_SIZE))
            self.cache += s
            self.cache_length += len(s)

        # Get the end of the requested block.

        next_start = self.cache_start + n
        s = self.cache[self.cache_start:next_start]

        # Reposition the pointer to the cache.

        self._seek_cache(len(s))
        return s

    def tell(self):
        return self.f.tell() - self.cache_length + self.cache_start

    def seek(self, offset):
        current = self.tell()
        self.f.seek(offset)

        # If seeking forward, attempt to navigate the cache.

        if offset >= current:
            self._seek_cache(offset - current)
        else:
            self.reset_cache()

    def _seek_cache(self, delta):
        next_start = self.cache_start + delta

        if next_start > 0 and next_start >= len(self.cache):
            self.reset_cache()

        # If the cache is too big, resize it.

        elif next_start > READ_CACHE_RESIZE:
            self.cache = self.cache[next_start:]
            self.cache_length = len(self.cache)
            self.cache_start = 0

        # Otherwise, just reference the next part of the cache.

        else:
            self.cache_start = next_start

class FileOpener:

    "Opening files using their filenames."

    def __init__(self, filename):
        self.filename = filename

    def open(self, mode):
        return open(self.filename, mode)

    def close(self):
        pass

# vim: tabstop=4 expandtab shiftwidth=4
