#!/usr/bin/env python

"""
Specific classes for storing position information.

Copyright (C) 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.files import *
from iixr.data import vint

class PositionWriter(FileWriter):

    "Writing position information to files."

    def reset(self):
        self.last_docnum = 0

    def write_positions(self, docnum, positions):

        """
        Write for the document 'docnum' the given 'positions'.
        Return the offset of the written record.
        """

        if docnum < self.last_docnum:
            raise ValueError, "Document number %r is less than previous number %r." % (docnum, self.last_docnum)

        # Record the offset of this record.

        offset = self.tell()

        # Make sure that the positions are sorted.

        positions.sort()

        # Write the position deltas.

        output = []
        last = 0

        for position in positions:
            output.append(vint(position - last))
            last = position

        # Write the document number delta.
        # Write the number of positions.
        # Then write the positions.

        self.write(vint(docnum - self.last_docnum) + vint(len(positions)) + "".join(output))

        self.last_docnum = docnum
        return offset

class PositionOpener(FileOpener):

    "Reading position information from files."

    def read_term_positions(self, offset, count):

        """
        Read all positions from 'offset', seeking to that position in the file
        before reading. The number of documents available for reading is limited
        to 'count'.
        """

        # Duplicate the file handle.

        f = self.open("rb")
        return PositionIterator(f, offset, count)

class PositionIndexWriter(FileWriter):

    "Writing position index information to files."

    def reset(self):
        self.last_docnum = 0
        self.last_pos_offset = 0

    def write_positions(self, docnum, pos_offset, count):

        """
        Write the given 'docnum, 'pos_offset' and document 'count' to the
        position index file.
        """

        # Record the offset of this record.

        offset = self.tell()
        output = []

        # Write the document number delta.

        output.append(vint(docnum - self.last_docnum))
        self.last_docnum = docnum

        # Write the position file offset delta.

        output.append(vint(pos_offset - self.last_pos_offset))
        self.last_pos_offset = pos_offset

        # Write the document count.

        output.append(vint(count))

        # Actually write the data.

        self.write("".join(output))

        return offset

class PositionIndexOpener(FileOpener):

    "Reading position index information from files."

    def read_term_positions(self, offset, doc_frequency):

        """
        Read all positions from 'offset', seeking to that position in the file
        before reading. The number of documents available for reading is limited
        to 'doc_frequency'.
        """

        # Duplicate the file handle.

        f = self.open("rb")
        return PositionIndexIterator(f, offset, doc_frequency)

# Iterators for position-related files.

class IteratorBase:

    def __init__(self, count):
        self.replenish(count)

    def replenish(self, count):
        self.count = count
        self.read_documents = 0

    def __len__(self):
        return self.count

    def sort(self):
        pass # Stored document positions are already sorted.

    def __iter__(self):
        return self

class PositionIterator(FileReader, IteratorBase):

    "Iterating over document positions."

    def __init__(self, f, offset, count):
        FileReader.__init__(self, f)
        IteratorBase.__init__(self, count)
        self.seek(offset)

    def reset(self):
        self.last_docnum = 0

    def read_positions(self):

        "Read positions, returning a document number and a list of positions."

        # Read the document number delta and add it to the last number.

        self.last_docnum += self.read_number()

        # Read the number of positions.

        npositions = self.read_number()

        # Read the position deltas, adding each previous position to get the
        # appropriate collection of absolute positions.

        i = 0
        last = 0
        positions = []

        while i < npositions:
            last += self.read_number()
            positions.append(last)
            i += 1

        return self.last_docnum, positions

    def next(self):

        "Read positions for a single document."

        if self.read_documents < self.count:
            self.read_documents += 1
            return self.read_positions()
        else:
            raise StopIteration

class PositionIndexIterator(FileReader, IteratorBase):

    "Iterating over document positions."

    def __init__(self, f, offset, count):
        FileReader.__init__(self, f)
        IteratorBase.__init__(self, count)
        self.seek(offset)
        self.section_count = 0

    def reset(self):
        self.last_docnum = 0
        self.last_pos_offset = 0

    def read_positions(self):

        """
        Read a document number, a position file offset for the position index
        file, and the number of documents in a section of that file.
        """

        # Read the document number delta.

        self.last_docnum += self.read_number()

        # Read the offset delta.

        self.last_pos_offset += self.read_number()

        # Read the document count.

        count = self.read_number()

        return self.last_docnum, self.last_pos_offset, count

    def next(self):

        "Read positions for a single document."

        self.read_documents += self.section_count
        if self.read_documents < self.count:
            docnum, pos_offset, self.section_count = t = self.read_positions()
            return t
        else:
            raise StopIteration

class PositionDictionaryWriter:

    "Writing position dictionaries."

    def __init__(self, position_writer, position_index_writer, interval):
        self.position_writer = position_writer
        self.position_index_writer = position_index_writer
        self.interval = interval

    def write_term_positions(self, doc_positions):

        """
        Write all 'doc_positions' - a collection of tuples of the form (document
        number, position list) - to the file.

        Add some records to the index, making dictionary entries.

        Return a tuple containing the offset of the written data, the frequency
        (number of positions), and document frequency (number of documents) for
        the term involved.
        """

        # Reset the writers.

        self.position_writer.reset()
        self.position_index_writer.reset()

        index_offset = None

        # Write the positions.

        frequency = 0
        first_docnum = None
        first_offset = None
        count = 0

        doc_positions.sort()

        for docnum, positions in doc_positions:
            pos_offset = self.position_writer.write_positions(docnum, positions)

            # Retain the first record offset for a subsequent index entry.

            if first_offset is None:
                first_offset = pos_offset
                first_docnum = docnum

            frequency += len(positions)
            count += 1

            # Every {interval} entries, write an index entry.

            if count % self.interval == 0:
                io = self.position_index_writer.write_positions(first_docnum, first_offset, self.interval)

                # Remember the first index entry offset.

                if index_offset is None:
                    index_offset = io

                first_offset = None
                first_docnum = None

                # Reset the position writer so that position readers accessing
                # a section start with the correct document number.

                self.position_writer.reset()

        # Finish writing an index entry for the remaining documents.

        else:
            if first_offset is not None:
                io = self.position_index_writer.write_positions(first_docnum, first_offset, count % self.interval)

                # Remember the first index entry offset.

                if index_offset is None:
                    index_offset = io

        return index_offset, frequency, count

    def close(self):
        self.position_writer.close()
        self.position_index_writer.close()

class PositionDictionaryReader:

    "Reading position dictionaries."

    def __init__(self, position_opener, position_index_opener):
        self.position_opener = position_opener
        self.position_index_opener = position_index_opener

    def read_term_positions(self, offset, doc_frequency):

        """
        Return an iterator for dictionary entries starting at 'offset' with the
        given 'doc_frequency'.
        """

        return PositionDictionaryIterator(self.position_opener,
            self.position_index_opener, offset, doc_frequency)

    def close(self):
        pass

class PositionDictionaryIterator:

    "Iteration over position dictionary entries."

    def __init__(self, position_opener, position_index_opener, offset, doc_frequency):
        self.position_opener = position_opener
        self.doc_frequency = doc_frequency
        self.index_iterator = position_index_opener.read_term_positions(offset, doc_frequency)
        self.iterator = None

        # Remember the last values.

        self.found_docnum, self.found_positions = None, None

        # Maintain state for the next index entry, if read.

        self.next_docnum, self.next_pos_offset, self.next_section_count = None, None, None

        # Initialise the current index entry and current position file iterator.

        self._next_section()
        self._init_section()

    # Sequence methods.

    def __len__(self):
        return self.doc_frequency

    def sort(self):
        pass

    # Iterator methods.

    def __iter__(self):
        return self

    def next(self):

        """
        Attempt to get the next document record from the section in the
        positions file.
        """

        # Return any visited but unrequested record.

        if self.found_docnum is not None:
            t = self.found_docnum, self.found_positions
            self.found_docnum, self.found_positions = None, None
            return t

        # Or search for the next record.

        while 1:

            # Either return the next record.

            try:
                return self.iterator.next()

            # Or, where a section is finished, get the next section and try again.

            except StopIteration:

                # Where a section follows, update the index iterator, but keep
                # reading using the same file iterator (since the data should
                # just follow on from the last section).

                self._next_section()
                self.iterator.replenish(self.section_count)

                # Reset the state of the iterator to make sure that document
                # numbers are correct.

                self.iterator.reset()

    def from_document(self, docnum):

        """
        Attempt to navigate to a positions entry for the given 'docnum',
        returning the positions for 'docnum', or None otherwise.
        """

        # Return any unrequested document positions.

        if docnum == self.found_docnum:
            return self.found_positions

        # Read ahead in the index until the next entry refers to a document
        # later than the desired document.

        try:
            if self.next_docnum is None:
                self.next_docnum, self.next_pos_offset, self.next_section_count = self.index_iterator.next()

            # Read until the next entry is after the desired document number,
            # or until the end of the results.

            while self.next_docnum <= docnum:
                self._next_read_section()
                if self.docnum < docnum:
                    self.next_docnum, self.next_pos_offset, self.next_section_count = self.index_iterator.next()
                else:
                    break

        except StopIteration:
            pass

        # Navigate in the position file to the document.

        self._init_section()

        try:
            while 1:
                found_docnum, found_positions = self.iterator.next()

                # Return the desired document positions or None (retaining the
                # positions for the document immediately after).

                if docnum == found_docnum:
                    return found_positions
                elif docnum < found_docnum:
                    self.found_docnum, self.found_positions = found_docnum, found_positions
                    return None

        except StopIteration:
            return None

    # Internal methods.

    def _next_section(self):

        "Attempt to get the next section in the index."

        if self.next_docnum is None:
            self.docnum, self.pos_offset, self.section_count = self.index_iterator.next()
        else:
            self._next_read_section()

    def _next_read_section(self):

        """
        Make the next index entry the current one without reading from the
        index.
        """

        self.docnum, self.pos_offset, self.section_count = self.next_docnum, self.next_pos_offset, self.next_section_count
        self.next_docnum, self.next_pos_offset, self.next_section_count = None, None, None

    def _init_section(self):

        "Initialise the iterator for the section in the position file."

        if self.iterator is not None:
            self.iterator.close()
        self.iterator = self.position_opener.read_term_positions(self.pos_offset, self.section_count)

    def close(self):
        if self.iterator is not None:
            self.iterator.close()
            self.iterator = None
        if self.index_iterator is not None:
            self.index_iterator.close()
            self.index_iterator = None

# vim: tabstop=4 expandtab shiftwidth=4
